#include "X3RichTextPrinter.h"

// Link against riched20.lib
#pragma comment(lib,"riched20.lib")

// The IID in riched20.lib in the Plaform SDK appears to be incorrect.
// This one is from the Microsoft Knowledge Base, article Q270161.
const IID IID_ITextServices = { 0x8d33f740, 0xcf58, 0x11ce, {0xa8, 0x9d, 0x00, 0xaa, 0x00, 0x6c, 0xad, 0xc5} };
const IID IID_ITextDocument = { 0x8CC497C0, 0xA1DF, 0x11CE, {0x80, 0x98, 0x00, 0xAA, 0x00, 0x47, 0xBE, 0x5D} };

X3RichTextPrinter::X3RichTextPrinter()
{
	// Create a default character format
	this->CreateDefaultCharFormat();

	// Create a default paragraph format
	this->CreateDefaultParaFormat();

	// Get an interface to the windowless rich edit control
	IUnknown* pUnk;
	if (CreateTextServices(NULL, (ITextHost*) this, &pUnk) == S_OK) {
		// Get ITextServices and ITextDocument interfaces
		pUnk->QueryInterface(IID_ITextDocument, (void **) &m_pTextDoc);
		pUnk->QueryInterface(IID_ITextServices, (void **) &m_pTextServ);
		pUnk->Release();
	}

	// Set initial ref count to 1
	m_lRefCnt = 1;
}

X3RichTextPrinter::~X3RichTextPrinter()
{
	//Handled by manual call to Release()
}

void X3RichTextPrinter::CreateDefaultCharFormat()
{
	HDC hDCScr;
	int nFontHei, nLogPxY;

	hDCScr = GetDC(NULL);
	// Get font height per MSDN
	nFontHei = -MulDiv(14, GetDeviceCaps(hDCScr, LOGPIXELSY), 72);
	// Get amount of pixels per vertical inch
	nLogPxY = GetDeviceCaps(hDCScr, LOGPIXELSY);
	ReleaseDC(NULL, hDCScr);

	memset(&m_sCharFormat, NULL, sizeof(m_sCharFormat));
	m_sCharFormat.cbSize = sizeof(m_sCharFormat);
	m_sCharFormat.dwMask = CFM_ALL;
	m_sCharFormat.dwEffects = NULL;
	// Convert font height to twips
	m_sCharFormat.yHeight = (-nFontHei) * (1440 / nLogPxY); 
	m_sCharFormat.yOffset = 0;
	// Default text colot is black
	m_sCharFormat.crTextColor = RGB(0, 0, 0);
	// Default font is Arial
	wcscpy(m_sCharFormat.szFaceName, L"Arial");
	m_sCharFormat.bPitchAndFamily = DEFAULT_PITCH;
	m_sCharFormat.bCharSet = ANSI_CHARSET;
}

void X3RichTextPrinter::CreateDefaultParaFormat()
{
	// Create a simple default paragraph format
	memset(&m_sParaFormat, NULL, sizeof(m_sParaFormat));
	m_sParaFormat.cbSize = sizeof(m_sParaFormat);
	m_sParaFormat.dwMask = PFM_ALL;
	m_sParaFormat.wAlignment = PFA_LEFT;
}

static DWORD CALLBACK X3EditStreamCallbackOUT(DWORD dwCookie, LPBYTE pbBuff, LONG cb, LONG *pcb)
{
	// Read out RTF text from the RichEdit Control into a stream
	// Used IStream since it maintains a RW pointer
	ULONG    lBytesWritten = 0;
	IStream* pStream = (IStream*) dwCookie;

	// Write RTF Text to stream and adjust by the amount of bytes written
	pStream->Write(pbBuff, (ULONG) cb, &lBytesWritten);
	(*pcb) += lBytesWritten;

	return S_OK;
}

static DWORD CALLBACK X3EditStreamCallbackIN(DWORD dwCookie, LPBYTE pbBuff, LONG cb, LONG *pcb)
{
	// Write RTF text to the RichEdit Control from the stream
	// Used an IStream since it maintains a RW pointer
	ULONG    lBytesRead = 0;
	IStream* pStream = (IStream*) dwCookie;
	
	// Copy RTF text data from the stream into pBuff and adjust by bytes read
	pStream->Read(pbBuff, cb, &lBytesRead);
	(*pcb) += lBytesRead;

	return S_OK;
}

void X3RichTextPrinter::getRTFText(BOOL bSelectionOnly, IStream** ppStream, ULONG* ulStrLen)
{
	EDITSTREAM	  sEditStrm;
	WPARAM		  wpFormat;
	STATSTG		  sStreamInfo;
	LARGE_INTEGER lnStreamPos;

	wpFormat = SF_RTF | ((bSelectionOnly) ? SFF_SELECTION : 0);
	sEditStrm.dwCookie = (DWORD) (*ppStream);
	sEditStrm.dwError = ERROR_SUCCESS;
	sEditStrm.pfnCallback = X3EditStreamCallbackOUT;
 	
	// The EM_STREAMOUT message causes a rich edit control to pass its contents to an applicationdefined EditStreamCallback 
	// callback function. The callback function can then write the stream of data to a file or any other location that it chooses. 
	m_pTextServ->TxSendMessage(EM_STREAMOUT, wpFormat, (LPARAM) &sEditStrm, NULL);
	
	// Get the amount of bytes written to the Stream
	memset(&sStreamInfo, NULL, sizeof(sStreamInfo));
	(*ppStream)->Stat(&sStreamInfo, STATFLAG_DEFAULT);
	(*ulStrLen) = sStreamInfo.cbSize.LowPart;

	// SEEK back to the beginning of the stream for programmer sanity
	memset(&lnStreamPos, NULL, sizeof(lnStreamPos));
	lnStreamPos.QuadPart = 0;
	(*ppStream)->Seek(lnStreamPos, STREAM_SEEK_SET, NULL);
}

HRESULT X3RichTextPrinter::setRTFText(BOOL bSelectionOnly, LPSTR lpszRTFText)
{
	EDITSTREAM    sEditStrm;
	WPARAM	      wpFormat;
    IStream*	  pStream = NULL;
	ULONG		  ulBytesWritten = 0;
	ULONG		  ulStrLen = (ULONG) lstrlenA(lpszRTFText);
	LARGE_INTEGER lnStreamPos;
	
	// Mke sure there's something to write
	if (ulStrLen == 0) return S_OK;
	// Create a sizeable stream on global memory and write the RTF text into the stream
	if (CreateStreamOnHGlobal(NULL, TRUE, &pStream) != S_OK) return E_FAIL;
	pStream->Write((void *) lpszRTFText, ulStrLen, &ulBytesWritten);
	
	// Make sure all text was written
	if (ulBytesWritten < ulStrLen) {
		pStream->Release();
		return E_OUTOFMEMORY;
	}
	
	// Seek to the beginning of the stream
	memset(&lnStreamPos, NULL, sizeof(lnStreamPos));
	lnStreamPos.QuadPart = 0;
	if (pStream->Seek(lnStreamPos, STREAM_SEEK_SET, NULL) != S_OK) {
		pStream->Release();
		return E_FAIL;
	}

	wpFormat = SF_RTF | ((bSelectionOnly) ? SFF_SELECTION : 0);
	sEditStrm.dwCookie = (DWORD) pStream;
	sEditStrm.dwError = ERROR_SUCCESS;
	sEditStrm.pfnCallback = X3EditStreamCallbackIN;
	
	// Write in the RTF text
	m_pTextServ->TxSendMessage(EM_STREAMIN, wpFormat, (LPARAM) &sEditStrm, NULL);

	// Cleanup...
	pStream->Release();
	pStream = NULL;
	return S_OK;
}

HRESULT STDMETHODCALLTYPE X3RichTextPrinter::QueryInterface(REFIID riid, void** ppvObject)
{
	PVOID pThis = (PVOID) this;

	// Return relevant interface
	if (IsEqualIID(riid, IID_ITextDocument))
		ppvObject = (void **) &(m_pTextDoc);
	else if (IsEqualIID(riid, IID_ITextServices))
		ppvObject = (void **) &(m_pTextServ);
	else if ((IsEqualIID(riid, IID_ITextHost)) || (IsEqualIID(riid, IID_IUnknown)))
		ppvObject = (void **) &(pThis);
	else
		ppvObject = NULL;
	
	if (ppvObject) {
		((IUnknown*)(*ppvObject))->AddRef();
		return S_OK;
	}
	return E_NOINTERFACE;
}

ULONG STDMETHODCALLTYPE X3RichTextPrinter::AddRef()
{
	// Thread safe increment
    m_lRefCnt = InterlockedIncrement(&m_lRefCnt);
	return (ULONG) m_lRefCnt;
}

ULONG STDMETHODCALLTYPE X3RichTextPrinter::Release()
{
	// Thread safe decrement
	m_lRefCnt = InterlockedDecrement(&m_lRefCnt);
	// Check for no more references
	if (m_lRefCnt == 0) 
	{
		if (m_pTextDoc)  
		{
			m_pTextDoc->Release();
			m_pTextDoc = NULL;
		}
		if (m_pTextServ) 
		{
			m_pTextServ->Release();
			m_pTextServ = NULL;
		}
		delete this;
	}
	return (ULONG) m_lRefCnt;
}

void X3RichTextPrinter::SetText(LPCWSTR text)
{
	// Get a range for the whole of the text in the control
	ITextRange* range = NULL;
	BSTR bsText = NULL;

	m_pTextDoc->Range(0,0,&range);
	range->MoveEnd(tomStory,1,NULL);
	
	// Replace the whole text
	bsText = SysAllocString(text);
	range->SetText(bsText);
}

void X3RichTextPrinter::Range(long cpFirst, long cpLim, ITextRange** ppRange)
{
	// Get the given range of text in the control
	m_pTextDoc->Range(cpFirst,cpLim,ppRange);
}

void X3RichTextPrinter::SizeText(HDC hDC, RECT* rect)
{
	long  lHeight, lWidth;
	SIZEL szExtent = {-1, -1};

	lWidth = (rect->right - rect->left);
	lHeight = 0;
	
	// Constrains the text to the RECT rect
	// lWidth MUST NOT be 0, lHeight MUST be 0
	// On Out:
	// lWidth will have the value of the widest line
	// lHeight will have the height of the entire text
	m_pTextServ->TxGetNaturalSize(DVASPECT_CONTENT,hDC,NULL,NULL,TXTNS_FITTOCONTENT,&szExtent,&lWidth,&lHeight);
	rect->right = (rect->left + lWidth);
	rect->bottom = (rect->top + lHeight);
}

BOOL X3RichTextPrinter::DrawText(HDC hDC, RECT* rect)
{
	// Draw the text in the windowless control onto the given device context,
	// within the given bounding rectangle.
	RECTL rc = { rect->left, rect->top, rect->right, rect->bottom };
	HRESULT hr = m_pTextServ->TxDraw(DVASPECT_CONTENT,0,NULL,NULL,hDC,0,&rc,NULL,NULL,NULL,0,0);
	return (SUCCEEDED(hr));
}

HRESULT X3RichTextPrinter::SendMessage(UINT msg, WPARAM wparam, LPARAM lparam, LRESULT *plresult)
{
	// Pass on message to windowless rich text control
	return m_pTextServ->TxSendMessage(msg, wparam, lparam, plresult);
}

HDC X3RichTextPrinter::TxGetDC()
{
	return 0;
}

INT X3RichTextPrinter::TxReleaseDC(HDC hdc)
{
	return 1;
}

BOOL X3RichTextPrinter::TxShowScrollBar(INT fnBar, BOOL fShow)
{
	return FALSE;
}

BOOL X3RichTextPrinter::TxEnableScrollBar(INT fuSBFlags, INT fuArrowflags)
{
	return FALSE;
}

BOOL X3RichTextPrinter::TxSetScrollRange(INT fnBar, LONG nMinPos, INT nMaxPos, BOOL fRedraw)
{
	return FALSE;
}

BOOL X3RichTextPrinter::TxSetScrollPos(INT fnBar, INT nPos, BOOL fRedraw)
{
	return FALSE;
}

void X3RichTextPrinter::TxInvalidateRect(LPCRECT prc, BOOL fMode)
{
}

void X3RichTextPrinter::TxViewChange(BOOL fUpdate)
{
} 

BOOL X3RichTextPrinter::TxCreateCaret(HBITMAP hbmp, INT xWidth, INT yHeight)
{
	return FALSE;
}

BOOL X3RichTextPrinter::TxShowCaret(BOOL fShow)
{
	return FALSE;
}

BOOL X3RichTextPrinter::TxSetCaretPos(INT x, INT y)
{
	return FALSE;
}

BOOL X3RichTextPrinter::TxSetTimer(UINT idTimer, UINT uTimeout)
{
	return FALSE;
}

void X3RichTextPrinter::TxKillTimer(UINT idTimer)
{
}

void X3RichTextPrinter::TxScrollWindowEx(INT dx, INT dy, LPCRECT lprcScroll, LPCRECT lprcClip, HRGN hrgnUpdate, LPRECT lprcUpdate, UINT fuScroll)
{
}

void X3RichTextPrinter::TxSetCapture(BOOL fCapture)
{
}

void X3RichTextPrinter::TxSetFocus()
{
}

void X3RichTextPrinter::TxSetCursor(HCURSOR hcur, BOOL fText)
{
}

BOOL X3RichTextPrinter::TxScreenToClient(LPPOINT lppt)
{
	return FALSE;
}

BOOL X3RichTextPrinter::TxClientToScreen(LPPOINT lppt)
{
	return FALSE;
}

HRESULT X3RichTextPrinter::TxActivate(LONG* plOldState)
{
	return E_FAIL;
}

HRESULT X3RichTextPrinter::TxDeactivate(LONG lNewState)
{
	return E_FAIL;
}

HRESULT X3RichTextPrinter::TxGetClientRect(LPRECT prc)
{
	return E_FAIL;
}

HRESULT X3RichTextPrinter::TxGetViewInset(LPRECT prc)
{
	// Set zero sized margins
	prc->left = 0;
	prc->top = 0;
	prc->right = 0;
	prc->bottom = 0;
	return S_OK;
}

HRESULT X3RichTextPrinter::TxGetCharFormat(const CHARFORMATW **ppCF)
{
	// Return the default UNICODE character format set up in the constructor
	*ppCF = &(this->m_sCharFormat);
	return S_OK;
}

HRESULT X3RichTextPrinter::TxGetParaFormat(const PARAFORMAT **ppPF)
{
	// Return the default paragraph format set up in the constructor
	*ppPF = &(this->m_sParaFormat);
	return S_OK;
}

COLORREF X3RichTextPrinter::TxGetSysColor(int nIndex)
{
	// Pass requests for colours on to Windows
	return ::GetSysColor(nIndex);
}

HRESULT X3RichTextPrinter::TxGetBackStyle(TXTBACKSTYLE* pstyle)
{
	// Do not erase what is underneath the drawing area
	*pstyle = TXTBACK_TRANSPARENT;
	return S_OK;
}

HRESULT X3RichTextPrinter::TxGetMaxLength(DWORD *plength)
{
	// Set the maximum size of text to be arbitrarily large
	*plength = 1024*1024*16;
	return S_OK;
}

HRESULT X3RichTextPrinter::TxGetScrollBars(DWORD *pdwScrollBar)
{
	// Do not allow scrollbars
	*pdwScrollBar = 0;
	return S_OK;
}

HRESULT X3RichTextPrinter::TxGetPasswordChar(TCHAR *pch)
{
	return S_FALSE;
}

HRESULT X3RichTextPrinter::TxGetAcceleratorPos(LONG *pcp)
{
	// A character position of 1 indicates that no character should be underlined. 
	*pcp = -1;
	return S_OK;
}

HRESULT X3RichTextPrinter::TxGetExtent(LPSIZEL lpExtent)
{
	// Requests the native size of the control in HIMETRIC.
	return E_NOTIMPL;
}

HRESULT X3RichTextPrinter::OnTxCharFormatChange(const CHARFORMATW* pcf)
{
	// Set the default character format for the text host
	memcpy(&m_sCharFormat, pcf, pcf->cbSize);   
	return S_OK;
}

HRESULT X3RichTextPrinter::OnTxParaFormatChange(const PARAFORMAT* ppf)
{
	// Set the default paragraph format for the text host
	memcpy(&m_sParaFormat, ppf, ppf->cbSize); 
	return S_OK;
}

HRESULT X3RichTextPrinter::TxGetPropertyBits(DWORD dwMask, DWORD *pdwBits)
{
	// Set the windowless control as being multiple lines of wrapping rich text
	DWORD bits = TXTBIT_MULTILINE|TXTBIT_RICHTEXT|TXTBIT_WORDWRAP;
	(*pdwBits) = bits & dwMask;
	return S_OK;
}

HRESULT X3RichTextPrinter::TxNotify(DWORD iNotify, void *pv)
{
	// Claim to have handled the notifcation, even though we always ignore it
	return S_OK;
}

HIMC X3RichTextPrinter::TxImmGetContext()
{
	return 0;
}

void X3RichTextPrinter::TxImmReleaseContext(HIMC himc) 
{
}

HRESULT X3RichTextPrinter::TxGetSelectionBarWidth(LONG *lSelBarWidth)
{
	// No selection bar
	(*lSelBarWidth) = 0;
	return S_OK;
}

